#include "Node.h"

#include "Task.h"
#include "Workers/CpuWorker.h"
#include "Workers/GpuWorker.h"
#include <stdio.h>
#include <stdlib.h>

bool g_single_worker;
int Node::mpi_rank;
int Node::mpi_size;
WorkerPool Node::m_workers;

Node::Node(const int num_cpus, const int num_omps, int max_gpus)
{
    for (int i = 0; i < num_cpus; i++) {
        m_workers.AddWorker( new CpuWorker(num_omps) );
    }
    
    const int avaliable_gpus = GpuWorker::GetNumberOfDevices();
    if (avaliable_gpus < max_gpus) {
        max_gpus = avaliable_gpus;
    }
    for (int i = 0; i < max_gpus; i++) {
        m_workers.AddWorker( new GpuWorker );
    }
    
    m_workers.MeasurePerformance();
    
    int provided;
    const int rc = MPI_Init_thread(NULL, NULL, MPI_THREAD_FUNNELED, &provided);
    if (rc != MPI_SUCCESS)
    {
        printf("ERROR: MPI_Init_thread failed with return code = %d.\n", rc);
        exit(EXIT_FAILURE);
    }

    MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
    
    g_single_worker = (mpi_size == 1 && num_cpus+max_gpus == 1);
}

bool Node::BroadcastEndSignal(bool end)
{
    MPI_Bcast(&end, sizeof(bool), MPI_BYTE, 0, MPI_COMM_WORLD);
    
    if (end == true)
    {
        const int rc = MPI_Finalize();    
        if (rc != MPI_SUCCESS)
        {
            printf("ERROR: MPI_Finalize() failed with return code = %d.\n", rc);
            exit(EXIT_FAILURE);
        }
        
        m_workers.KillAllWorkers();
    }
    
    return end;
}

void Node::RunAsSlave()
{
    while (true)
    {
        const bool end = BroadcastEndSignal();
        if (end == true) {
            break;
        }
    
        /* Receive the Task's metadata from the host Node */
        Task *task = (Task*) malloc(sizeof(Task));
        MPI_Bcast(task, sizeof(Task), MPI_BYTE, 0, MPI_COMM_WORLD);

        int local_x_size = task->m_num_x / mpi_size;
        const int x_remainder = task->m_num_x % mpi_size;
        int from_x = mpi_rank * local_x_size;
        
        if (mpi_rank < x_remainder) 
        {
            local_x_size++;
            from_x += mpi_rank;
        }                
        else {
            from_x += x_remainder;
        }                          

        /* Receive the subproblem from the host Node; containing the electromagnetic field values */
        MPI_Status status;
        MPI_Probe(0, FIELD_DATA_TAG, MPI_COMM_WORLD, &status);
        int buffer_size;        
        MPI_Get_count(&status, MPI_BYTE, &buffer_size);    
        char *buffer = (char*) malloc(buffer_size);
        MPI_Recv(buffer, buffer_size, MPI_BYTE, 0, FIELD_DATA_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        task->UpdateFieldPointers(local_x_size, buffer);
        
        const int to_x = from_x + local_x_size;
        m_workers.PrepareForNewTask(from_x, to_x, task);                

        MPI_Datatype up_e_datatype, down_e_datatype;
        MPI_Datatype up_h_datatype, down_h_datatype;
        CreateE_BordersDatatypes(local_x_size, *task, &up_e_datatype, &down_e_datatype);
        CreateH_BordersDatatypes(local_x_size, *task, &up_h_datatype, &down_h_datatype);
    
        /* Solve the FDTD calculations, while doing the necessary border exchanges between updates */
        for (int time_step = 1; time_step <= task->m_num_time_steps; time_step++)
        {       
            m_workers.UpdateE();
            ExchangeE_Borders(up_e_datatype, down_e_datatype, task);        
    
            m_workers.SourceExcitation(time_step, task->m_is);
            
            m_workers.UpdateH();
            ExchangeH_Borders(up_h_datatype, down_h_datatype, task);                                    
        }
        m_workers.FinalizeTask();
        
        /* Send the completed subproblem back to the host Node */
        MPI_Send(buffer, buffer_size, MPI_BYTE, 0, FIELD_DATA_TAG, MPI_COMM_WORLD);
        
        MPI_Type_free(&up_e_datatype);
        MPI_Type_free(&up_h_datatype);
        MPI_Type_free(&down_e_datatype);
        MPI_Type_free(&down_h_datatype);        
    
        free(buffer);
        free(task);    
    }
}

void Node::ExecuteTask(Task* task)
{
    task->m_totalTimer.Start();
    task->m_setupTimer.Start();    

    BroadcastEndSignal(false);
    
    /* Broadcast the Task's metadata to the slave Nodes */
    MPI_Bcast(task, sizeof(Task), MPI_BYTE, 0, MPI_COMM_WORLD);

    int local_x_size = task->m_num_x / mpi_size;
    const int x_remainder = task->m_num_x % mpi_size;
    int from_x = mpi_rank * local_x_size;

    if (mpi_rank < x_remainder) 
    {
        local_x_size++;
        from_x += mpi_rank;
    }                
    else {
        from_x += x_remainder;
    }        

    /* Split the problem into subproblems and distributes it evenly among the slave Nodes */
    std::vector<MPI_Datatype> datatypes;    
    datatypes.resize(mpi_size-1);
    for (int rank = 1; rank < mpi_size; rank++)
    {    
        CreateDatatypeFromTask(rank, *task, &datatypes[rank-1]);
        MPI_Send(task, 1, datatypes[rank-1], rank, FIELD_DATA_TAG, MPI_COMM_WORLD);    
    }        

    const int to_x = from_x + local_x_size;
    m_workers.PrepareForNewTask(from_x, to_x, task);

    MPI_Datatype up_e_datatype, down_e_datatype;
    MPI_Datatype up_h_datatype, down_h_datatype;
    CreateE_BordersDatatypes(local_x_size, *task, &up_e_datatype, &down_e_datatype);
    CreateH_BordersDatatypes(local_x_size, *task, &up_h_datatype, &down_h_datatype);

    task->m_setupTimer.Stop();
    task->m_computeTimer.Start();
    
    /* Solve the FDTD calculations, while doing the necessary border exchanges between updates */
    for (int time_step = 1; time_step <= task->m_num_time_steps; time_step++)
    {        
        m_workers.UpdateE();
        ExchangeE_Borders(up_e_datatype, down_e_datatype, task);        
    
        m_workers.SourceExcitation(time_step, task->m_is);
        
        m_workers.UpdateH();
        ExchangeH_Borders(up_h_datatype, down_h_datatype, task);                
    }  
    task->m_computeTimer.Stop();   
    m_workers.FinalizeTask();

    /* Gather all the completed subproblems from the slave Nodes */
    for (int rank = 1; rank < mpi_size; rank++)
    {
        MPI_Recv(task, 1, datatypes[rank-1], rank, FIELD_DATA_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        MPI_Type_free(&datatypes[rank-1]);    
    }
    
    MPI_Type_free(&up_e_datatype);
    MPI_Type_free(&up_h_datatype);
    MPI_Type_free(&down_e_datatype);
    MPI_Type_free(&down_h_datatype);            

    task->m_totalTimer.Stop();
    task->m_mflops_performance = task->m_flops_work / (task->m_totalTimer.GetTime()*1000*1000);
    task->m_done = true;        
}

void Node::CreateDatatypeFromTask(const int slave_rank, const Task &task, MPI_Datatype *datatype)
{
    const int yz_offset_e = (task.m_num_y+1) * (task.m_num_z+1);         
    const int yz_offset_h = task.m_num_y * task.m_num_z; 

    int rank_x_size = task.m_num_x / mpi_size;
    int e_displacement = rank_x_size * yz_offset_e * slave_rank;
    int h_displacement = (rank_x_size-1) * yz_offset_h * slave_rank;
    
    /* Handles cases where work has NOT been evenly divided among Nodes */    
    const int x_remainder = task.m_num_x % mpi_size;
    if (slave_rank < x_remainder) {
        rank_x_size++;
    }    
    for (int rank = 0; rank < x_remainder && rank < slave_rank; rank++)
    {
        e_displacement += yz_offset_e;
        h_displacement += yz_offset_h;
    }
    
    const int e_size = (rank_x_size+1) * yz_offset_e * sizeof(my_float);
    const int h_size = (rank_x_size+1) * yz_offset_h * sizeof(my_float);

    const int number_of_blocks = 6;
    
    int block_lengths[number_of_blocks];      
    block_lengths[0] = e_size;
    block_lengths[1] = e_size;
    block_lengths[2] = e_size;
    block_lengths[3] = h_size;
    block_lengths[4] = h_size;
    block_lengths[5] = h_size;

    MPI_Aint displacements[number_of_blocks]; 
    displacements[0] = (MPI_Aint)&task.m_ex[e_displacement] - (MPI_Aint)&task;           
    displacements[1] = (MPI_Aint)&task.m_ey[e_displacement] - (MPI_Aint)&task;     
    displacements[2] = (MPI_Aint)&task.m_ez[e_displacement] - (MPI_Aint)&task;      
    displacements[3] = (MPI_Aint)&task.m_hx[h_displacement] - (MPI_Aint)&task;
    displacements[4] = (MPI_Aint)&task.m_hy[h_displacement] - (MPI_Aint)&task;
    displacements[5] = (MPI_Aint)&task.m_hz[h_displacement] - (MPI_Aint)&task;    
    
    MPI_Datatype types[number_of_blocks];    
    std::fill_n(types, number_of_blocks, MPI_BYTE); 

    MPI_Type_create_struct(number_of_blocks, block_lengths, displacements, types, datatype);
    MPI_Type_commit(datatype);
}

void Node::CreateH_BordersDatatypes(const int local_x_size, const Task &task, 
    MPI_Datatype *up_h_datatype, MPI_Datatype *down_h_datatype)    
{
    const int number_of_blocks = 2;    
    int block_lengths[number_of_blocks];
    MPI_Aint displacements[number_of_blocks];
    MPI_Datatype types[number_of_blocks];    
    std::fill_n(types, number_of_blocks, MPI_BYTE);
    
    const int yz_offset_h = task.m_num_y * task.m_num_z;    
    const int h_size = yz_offset_h * sizeof(my_float);    
    std::fill_n(block_lengths, number_of_blocks, h_size);

    displacements[0] = (MPI_Aint)&task.m_hy[(local_x_size-1)*yz_offset_h] - (MPI_Aint)&task;
    displacements[1] = (MPI_Aint)&task.m_hz[(local_x_size-1)*yz_offset_h] - (MPI_Aint)&task;
    
    MPI_Type_create_struct(number_of_blocks, block_lengths, displacements, types, up_h_datatype);
    MPI_Type_commit(up_h_datatype);    
    
    displacements[0] = (MPI_Aint)&task.m_hy[-yz_offset_h] - (MPI_Aint)&task;
    displacements[1] = (MPI_Aint)&task.m_hz[-yz_offset_h] - (MPI_Aint)&task;    
    
    MPI_Type_create_struct(number_of_blocks, block_lengths, displacements, types, down_h_datatype);
    MPI_Type_commit(down_h_datatype);    
}

void Node::CreateE_BordersDatatypes(const int local_x_size, const Task &task, 
    MPI_Datatype *up_e_datatype, MPI_Datatype *down_e_datatype)
{       
    const int number_of_blocks = 2;    
    int block_lengths[number_of_blocks];
    MPI_Aint displacements[number_of_blocks];
    MPI_Datatype types[number_of_blocks];    
    std::fill_n(types, number_of_blocks, MPI_BYTE);

    const int yz_offset_e = (task.m_num_y+1) * (task.m_num_z+1);    
    const int e_size = yz_offset_e * sizeof(my_float);            
    std::fill_n(block_lengths, number_of_blocks, e_size);
 
    displacements[0] = (MPI_Aint)&task.m_ey[local_x_size*yz_offset_e] - (MPI_Aint)&task;
    displacements[1] = (MPI_Aint)&task.m_ez[local_x_size*yz_offset_e] - (MPI_Aint)&task;               
     
     MPI_Type_create_struct(number_of_blocks, block_lengths, displacements, types, up_e_datatype);
    MPI_Type_commit(up_e_datatype);
    
    displacements[0] = (MPI_Aint)&task.m_ey[0] - (MPI_Aint)&task;
    displacements[1] = (MPI_Aint)&task.m_ez[0] - (MPI_Aint)&task;
    
    MPI_Type_create_struct(number_of_blocks, block_lengths, displacements, types, down_e_datatype);
    MPI_Type_commit(down_e_datatype);        
}

void Node::ExchangeH_Borders(const MPI_Datatype up_h_datatype, const MPI_Datatype down_h_datatype, Task *task)
{
    if (mpi_size == 1)
        return;

    int rc;

    if (mpi_rank == 0) {
        rc = MPI_Send(task, 1, up_h_datatype, mpi_rank+1, FIELD_BORDER_TAG, MPI_COMM_WORLD);        
    }
    else if (mpi_rank == mpi_size-1) {
        rc = MPI_Recv(task, 1, down_h_datatype, mpi_rank-1, FIELD_BORDER_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }
    else {
        rc = MPI_Sendrecv(task, 1, up_h_datatype, mpi_rank+1, FIELD_BORDER_TAG,
            task, 1, down_h_datatype, mpi_rank-1, FIELD_BORDER_TAG, 
            MPI_COMM_WORLD, MPI_STATUS_IGNORE); 
    }
    
    if (rc != MPI_SUCCESS)
    {
        printf("ERROR: Node %d failed in ManageHBorders() with return code = %d.\n", mpi_rank, rc);
        exit(EXIT_FAILURE);
    }    
}

void Node::ExchangeE_Borders(const MPI_Datatype up_e_datatype, const MPI_Datatype down_e_datatype, Task *task)
{
    if (mpi_size == 1)
        return;

    int rc;
    
    if (mpi_rank == 0) {        
        rc = MPI_Recv(task, 1, up_e_datatype, mpi_rank+1, FIELD_BORDER_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }
    else if (mpi_rank == mpi_size-1) {        
        rc = MPI_Send(task, 1, down_e_datatype, mpi_rank-1, FIELD_BORDER_TAG, MPI_COMM_WORLD);
    }
    else {
        rc = MPI_Sendrecv(task, 1, down_e_datatype, mpi_rank-1, FIELD_BORDER_TAG,
            task, 1, up_e_datatype, mpi_rank+1, FIELD_BORDER_TAG, 
            MPI_COMM_WORLD, MPI_STATUS_IGNORE); 
    }
    
    if (rc != MPI_SUCCESS)
    {
        printf("ERROR: Node %d failed in ManageEBorders() with return code = %d.\n", mpi_rank, rc);
        exit(EXIT_FAILURE);
    }
}
