#ifndef NODE_H_
#define NODE_H_

#include "WorkerPool.h"
#include <mpi.h>

class Task;

/// An all static class that provides all the MPI functionality of the program.
/*! This purpose of this class is to serve as an utility class, which in this case means that it will handle the top level control of the program through the use of the rest of the implemented classes. This class will also contain all the program’s functions that relies on MPI. The class is implemented as an all static class in order to stress that there should only be one of it (per node). Using an all static class is a subjective choice, as it could also have been implemented as a namespace or as a singleton class. */
class Node
{
public:
/// Constructor which also creates workers according to the parameters.
/*! This function will create Workers according to the input parameters, store the workers in #m_workers, and measure their performances by calling WorkerPool::MeasurePerformance(). The function will also initialize MPI by calling MPI_Initialize, and should therefore only be called once\!
 *  @param[in] num_cpus The number of CPU workers to create.
 *  @param[in] num_omps The number of OpenMP threads each CPU worker will have.
 *  @param[in] max_gpus The maximum number of GPU workers to create. Will be changed to the number of installed GPUs if a number higher than the number of installed GPUs is used. */    
    Node(const int num_cpus, const int num_omps, int max_gpus);
    
/// Executes a Task using all avaliable Nodes and Workers.
/*! This function will execute a Task using all the available nodes and workers, and should only be called by the host Node after it has been initialized with at least one Worker. Only the electromagnetic field values of the task object will be modified. 
 *  @param[in,out] task A pointer to the task that will be executed. */    
    static void ExecuteTask(Task* task);
    
/// Puts the Node in a slave state where it will wait for work from the host node.
/*! This function should be called by every node except from the host node. Calling this function puts the slave node in a wait state, which implies running a loop with an execution path similar to that of the ExecuteTask() function called by the host node. The slave node will remain in this loop, where it will either execute tasks or wait for new tasks, until the host node calls BroadcastEndSignal(). */
    static void RunAsSlave();
    
/// Signals all the slave nodes to exit their slave state.
/*! This function uses MPI to broadcast a boolean to all the slave nodes, telling them if they should exit their slave state or continue to wait for new tasks. Calling this function is the only way (except for errors) to stop nodes executing the RunAsSlave() function. If a \i true value is broadcasted; all the nodes (including the host node) will call MPI_Finalize() and terminate all their workers by calling WorkerPool::KillAllWorkers().
 *  @param[in] end The bool value that should be broadcasted (only matters if called by the host node) */    
    static bool BroadcastEndSignal(bool end = true);
        
    static int mpi_rank; ///< The MPI rank of the node. Is used to differentiate the nodes from each other. Rank 0 represents the host node.
    static int mpi_size; ///< The total number of MPI processes in the communicator. Also referred to as the total number of nodes.
    
private:
/// Tags used to differentiate the MPI messages. Is not really needed, but included for readability.
    enum MyMpiTag { FIELD_DATA_TAG, FIELD_BORDER_TAG };

/// Creates and commits a MPI_Datatype for the electromagnetic field values of a Task object.
/*! This function should only be called by the host node. It will create a MPI derived datatype representing the electromagnetic field values that should be handled by the slave node specified by \p slave_rank. It will use #mpi_size and the provided \p slave_rank argument to calculate a suitable sub-domain size for the slave node. The purpose of creating the derived datatype is to be able to combine six electromagnetic field arrays into one MPI_Send(), and to be able to later receive the completed sub-domain from the associated slave node. The function also commits the datatype, but it is left to the caller to later free it, when it is no longer needed (MPI_Type_free()). 
 *  @param[in] slave_rank The rank of the slave node that the datatype will be made for.
 *  @param[in] task The task object containing the electromagnetic field arrays that the datatype will be a associated with.
 *  @param[out] datatype A pointer to the datatype that will contain the created and commited datatype. */ 
    static void CreateDatatypeFromTask(const int slave_rank, const Task &task, MPI_Datatype *datatype);
    
/// Creates datatypes for the top and bottom boundary data of the magnetic field sub-domains. 
/*! This function creates two MPI derived datatypes: one for the top and one for the bottom border of the sub-domain’s magnetic field values. These datatypes are used to simplify the border exchanges between adjacent nodes, by bundling together borders of both the Y and Z directions. The datatypes are also committed, and should later be freed by the user (when no longer needed) by calling MPI_Type_free().
 *  @param[in] local_x_size The size in the X direction of the node's local sub-domain.
 *  @param[in] task The local copy of the task object.
 *  @param[out] up_h_datatype A pointer to the datatype that will contain the created and committed datatype for the uppermost (in the X direction) borders of the magnetic field values in the Y and Z directions.
 *  @param[out] down_h_datatype Pointer to the datatype that will contain the created and committed datatype for the lowest (in the X direction) borders of the magnetic field values in the Y and Z directions. */
    static void CreateH_BordersDatatypes(const int local_x_size, const Task &task, MPI_Datatype *up_h_datatype, MPI_Datatype *down_h_datatype);
    
/// Creates datatypes for the top and bottom boundary data of the electric field sub-domains. 
/*! This function creates two MPI derived datatypes: one for the top and one for the bottom border of the sub-domain’s electric field values. These datatypes are used to simplify the border exchanges between adjacent nodes, by bundling together borders of both the Y and Z directions. The datatypes are also committed, and should later be freed by the user (when no longer needed) by calling MPI_Type_free().
 *  @param[in] local_x_size The size in the x direction of the node's local subdomain.
 *  @param[in] task The local copy of the task object.
 *  @param[out] up_e_datatype Pointer to the datatype that will contain the created and committed datatype, for the uppermost (in the X direction) borders of the electric field values in the Y and Z directions.
 *  @param[out] down_e_datatype Pointer to the datatype that will contain the created and committed datatype, for the lowest (in the X direction) borders of the electric field values in the Y and Z directions. */    
    static void CreateE_BordersDatatypes(const int local_x_size, const Task &task, MPI_Datatype *up_e_datatype, MPI_Datatype *down_e_datatype);    
    
/// Exchanges the borders of the magnetic field component values with adjacent nodes in the X direction.
/*! CreateH_BordersDatatypes() should be called prior to this function, since it depends on the datatypes it creates.
 *  @param[in] up_h_datatype Datatype for the uppermost (in the X direction) borders of the magnetic field values in the Y and Z directions.
 *  @param[in] down_h_datatype Datatype for the lowest (in the X direction) borders of the magnetic field values in the Y and Z directions. 
 *  @param[in,out] task The local task object, used as a base address by the datatypes. */
    static void ExchangeH_Borders(const MPI_Datatype up_h_datatype, const MPI_Datatype down_h_datatype, Task *task);
    
/// Exchanges the borders of the electric field component values with adjacent nodes in the X direction.
/*! CreateE_BordersDatatypes() should be called prior to this function, since it depends on the datatypes it creates.
 *  @param[in] up_e_datatype Datatype for the uppermost (in the X direction) borders of the electric field values in the Y and Z directions.
 *  @param[in] down_e_datatype Datatype for the lowest (in the X direction) borders of the electric field values in the Y and Z directions. 
 *  @param[in,out] task The local task object, used as a base address by the datatypes. */    
    static void ExchangeE_Borders(const MPI_Datatype up_e_datatype, const MPI_Datatype down_e_datatype, Task *task);

    static WorkerPool m_workers; ///< Container for all the node's workers.
};

#endif /* NODE_H_ */
