#include "Task.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

int Task::s_next_id = 0;

Task::Task(int nts, int nx, int ny, int nz, int dx, int dy, int dz) 
    : m_done(false), m_id(s_next_id++), m_num_time_steps(nts), 
        m_num_x(nx), m_num_y(ny), m_num_z(nz),  
        m_bytes(0), m_mflops_performance(0.0)
{
    if (m_num_time_steps <= 0)
    {
        printf("ERROR: Time steps in Task %d is zero or less.\n", m_id);
        exit (EXIT_FAILURE);
    }
    if (m_num_x <= 0 || m_num_y <= 0 || m_num_z <= 0)
    {
        printf("ERROR: A dimension in Task %d is zero or less.\n", m_id);
        exit (EXIT_FAILURE);
    }

    /* Calculate the total Flops of the task */
    m_flops_work =  double(m_num_y-1)*(m_num_z-1)*6;
    m_flops_work += double(m_num_x-1)*(m_num_z-1)*6;
    m_flops_work += double(m_num_x-1)*(m_num_y-1)*6;
    m_flops_work += double(m_num_x-1)*(m_num_y-1)*(m_num_z-1)*18;
    m_flops_work += double(m_num_x)*m_num_y*m_num_z*18;
    m_flops_work *= m_num_time_steps;

    /* Initialize leap-frog update */
    m_dt = sqrt(3) / (2*s_c0*sqrt(1/(dx*dx) + 1/(dy*dy) + 1/(dz*dz)));
    m_Cbdx = m_dt/(s_mu0*dx);
    m_Cbdy = m_dt/(s_mu0*dy);
    m_Cbdz = m_dt/(s_mu0*dz);
    m_Dbdx = m_dt/(s_eps0*dx);
    m_Dbdy = m_dt/(s_eps0*dy);
    m_Dbdz = m_dt/(s_eps0*dz);

    /* Initialize point source */
    m_p_source_factor = m_dt / (s_eps0*dx*dy*dz) * s_const*2;
    m_is = (m_num_x/2) - 1;
    m_js = (m_num_y/2) - 1;
    m_ks = (m_num_z/2) - 1;
}

void Task::CleanUp()
{
    free(m_ex);
    free(m_ey);
    free(m_ez);
    free(m_hx);
    free(m_hy);
    free(m_hz);
}

void Task::InitializeFieldComponents()
{
    /* Calculate number of field components */
    const long long num_e_elements = (m_num_x+1)*(m_num_y+1)*(m_num_z+1);
    const long long num_h_elements = m_num_x*m_num_y*m_num_z;

    m_ex = (my_float*) calloc(num_e_elements, sizeof(my_float));
    m_ey = (my_float*) calloc(num_e_elements, sizeof(my_float));
    m_ez = (my_float*) calloc(num_e_elements, sizeof(my_float));
    m_hx = (my_float*) calloc(num_h_elements, sizeof(my_float));
    m_hy = (my_float*) calloc(num_h_elements, sizeof(my_float));
    m_hz = (my_float*) calloc(num_h_elements, sizeof(my_float));

    m_bytes = 3 * (num_e_elements + num_h_elements) * sizeof(my_float);

    if (g_output == VERBOSE)
    {
        printf("Electromagnetic fields initialized for Task %d:\n"
            " - Total size: %lld bytes (%lld megabytes)\n\n",
            m_id, m_bytes, m_bytes/(1024*1024));    
    }
}

void Task::UpdateFieldPointers(const int local_x_size, const char *buffer)
{   
    if (buffer == NULL)
        return;

    const int yz_offset_e_size = (m_num_y+1) * (m_num_z+1) * sizeof(my_float);
    const int yz_offset_h_size = m_num_y * m_num_z * sizeof(my_float);
    const int e_size = (local_x_size+1) * yz_offset_e_size;
    const int h_size = local_x_size * yz_offset_h_size ;    
    
    int offset = 0;    
    m_ex = (my_float*)&buffer[offset];
    offset += e_size;
    m_ey = (my_float*)&buffer[offset];
    offset += e_size;
    m_ez = (my_float*)&buffer[offset];
    offset += e_size + yz_offset_h_size;
    
    m_hx = (my_float*)&buffer[offset];
    offset += h_size + yz_offset_h_size;
    m_hy = (my_float*)&buffer[offset];
    offset += h_size + yz_offset_h_size;
    m_hz = (my_float*)&buffer[offset];
}

void Task::PrintResults()
{
    if (g_output != VERBOSE) {
        return;
    }

    const float mflops = m_flops_work / (1000*1000);
    const float total_time = m_totalTimer.GetTime();
    const float mflops_per_second = mflops / total_time;
        
    printf("Task %d completed:\n"
        " - Total MFLOPs of Task: %.3f\n"
        " - Total time: %.3f sec.\n"
        " - MFLOPs/s: %.3f\n\n",
        m_id, mflops, total_time, mflops_per_second);
    
    if (m_num_time_steps == 200 && m_num_x == 100 && m_num_y == 100 && m_num_z == 100)
    {
        const my_float expected_sum_Ez = -0.0692134;
        my_float sum_Ez = 0.0;    
        
        for (int i = 0; i < m_num_x; i++)
            for (int j = 0; j < m_num_y; j++)
                for (int k = 0; k < m_num_z; k++)
                    sum_Ez += m_ez[(m_num_y+1)*(m_num_z+1)*i + (m_num_z+1)*j + k];
                
        printf(" - Checksum(Ez)   = %25.17E\n"
            " - Expected value = %25.17E\n\n", 
            sum_Ez, expected_sum_Ez);    
    }    
}
