#ifndef TASK_H_
#define TASK_H_

#include "globals.h"
#include "Timer.h"

/// Container class for all the data associated with a FDTD domain.
/*! Each instance of the Task class represents an individual FDTD domain, and contains all the data associated with that domain. Executing a Task simulates the work required for a real FDTD simulation. Details about the FDTD domain, such as its size and Yee cell details, are specified by the user during object construction. The domains are currently limited to only using a homogeneous media, meaning that there is only a single material across the entire domain, this material is set to empty space (vacuum). All the electromagnetic fields are initialized to zero. A point source is located at the center of the domain.*/ 
class Task {
public:
/// Constructor.
/*! Creates an instance of the Task class using the specifications provided as arguments. Will generate an unique ID: #m_id, and calculate the total required FLOPs of the Task: #m_flops_work. Will also set the point source location: #m_is, #m_js, #m_ks, to the center of the domain. Calling this function will not allocate memory for the electromagnetic field values, it should therefore be followed by a call to: InitializeFieldComponents(), before doing calculations on the Task.
 *  @param[in] nts The number of time-steps.
 *  @param[in] nx,ny,nz Specifies the number of Yee-Cells in each dimension. Also known as the domain size or grid size.
 *  @param[in] dx,dy,dz The size of each Yee-Cell in each dimension @f$(\Delta x,\Delta y,\Delta z)@f$. */
    Task(int nts, int nx, int ny, int nz, int dx, int dy, int dz);
    
/// Dynamically allocates memory for the field components. 
/*! Will dynamically allocate the required memory for the electromagnetic field components, depending on the specified dimensions: #dx, #dy, #dz, and the precision of the values (32/64 bit). All the values will also be initialized to zero. Must be freed by calling CleanUp() before ~Task(). */
    void InitializeFieldComponents();
    
/// Updates the Task's pointers to point to the data stored in the local \p buffer. 
/*! This function is used when the Task has been received from another process\node. Because the pointers are transferred “as is”, they will still use the addresses in the sender’s memory space, which will result in undefined behavior if used by the receiver. Calling this function will update the pointers to point to data in the local memory space.
 *  @param[in] local_x_size The number of Yee-Cells in the X dimension of the local subdomain of the Task, will be used instead of the \a m_nx, and together with \a m_ny and \a m_nz, to calculate the offsets between the field component arrays in the \p buffer.
 *  @param[in] buffer A pointer to the start of the \p buffer, where all the local field components are stored contiguously in memory. */
    void UpdateFieldPointers(const int local_x_size, const char *buffer);    
    
/// Frees the memory allocated by InitializeFieldComponents(). Must be called before ~Task().
    void CleanUp();

/// Prints the Task's performance results. 
/*! Should only be called after a Task has been executed by a Worker. If the Task uses the test configuration of: 200 time steps, 100*100*100 Yee-Cells, and 1.0*1.0*1.0 Yee-Cell size. It will also print the sum of the #m_ez array and compare it to the known correct result of said test configuration. */
    void PrintResults();
    
    bool m_done; ///< A flag that indicates if the Task has been completed. 
    const int m_id; ///< An unique ID automatically generated during object construction. 
    
    const int m_num_time_steps; ///< The number of time steps. 
    const int m_num_x; ///< The number of Yee-Cells in the X dimension. 
    const int m_num_y; ///< The number of Yee-Cells in the Y dimension. 
    const int m_num_z; ///< The number of Yee-Cells in the Z dimension. 
    
    long long m_bytes; ///< The total number of bytes allocated for the field components. Will only be set after a call to InitializeFieldComponents(). 
    
    int m_is; ///< The X-coordinate of the source excitation, also known as the point source.
    int m_js; ///< The Y-coordinate of the source excitation, also known as the point source.
    int m_ks; ///< The Z-coordinate of the source excitation, also known as the point source.    
    
    my_float m_p_source_factor; ///< The point source factor.
    
    my_float m_dt; ///< The time of one time-step: @f$\Delta t @f$.
    my_float m_Cbdx; ///< Constant: @f$\frac{\Delta t}{\mu_0\Delta x}@f$ 
    my_float m_Cbdy; ///< Constant: @f$\frac{\Delta t}{\mu_0\Delta y}@f$ 
    my_float m_Cbdz; ///< Constant: @f$\frac{\Delta t}{\mu_0\Delta z}@f$ 
    my_float m_Dbdx; ///< Constant: @f$\frac{\Delta t}{\epsilon_0\Delta x}@f$ 
    my_float m_Dbdy; ///< Constant: @f$\frac{\Delta t}{\epsilon_0\Delta y}@f$ 
    my_float m_Dbdz; ///< Constant: @f$\frac{\Delta t}{\epsilon_0\Delta z}@f$     
    
    Timer m_setupTimer; ///< Contains the time (in seconds) used by the program to get the Task ready for execution.
    Timer m_computeTimer; ///< Contains the time (in seconds) used for the required computations of the Task.
    Timer m_totalTimer; ///< Contains the total time (in seconds) used to complete the Task.
    
    double m_flops_work; ///< The number of floating-point operations required to complete the Task. 
    double m_mflops_performance; ///< The number of million floating-point operations per second the Task was executed with. 

    my_float *m_ex; ///< Electrical field components in X direction. 
    my_float *m_ey; ///< Electrical field components in Y direction.
    my_float *m_ez; ///< Electrical field components in Z direction.
    my_float *m_hx; ///< Magnetic field components in X direction.
    my_float *m_hy; ///< Magnetic field components in Y direction.
    my_float *m_hz; ///< Magnetic field components in Z direction.

private:
    static int s_next_id; ///< A static counter used to generate the unique #m_id. 
    static const my_float s_const = 1.0e-10; ///< A constant used to set the #m_p_source_factor.

/*! @brief Relative permeability of vacuum: @f$\mu_0 @f$.
 *  The degree of magnetization that a material obtains in response to an applied magnetic field. */
    static const my_float s_mu0 = 1.256637061E-6;

/*! @brief Relative Permittivity of vacuum: @f$\epsilon_0 @f$.
 *  The ability of a substance to store electrical energy in an electric field. */
    static const my_float s_eps0 = 8.8541878E-12;

    static const my_float s_c0 = 2.99792458E+8; ///< The speed of light in vacuum.
};

#endif /* TASK_H_ */

