#include "WorkerPool.h"

#include "globals.h"
#include "Task.h"
#include "Workers/Worker.h"
#include <stdio.h>

void WorkerPool::AddWorker(Worker *new_worker)
{
    m_workers.push_back(new_worker);
}

void WorkerPool::MeasurePerformance()
{    
    for (unsigned int i = 0; i < m_workers.size(); i++) {
        m_workers[i]->MeasurePerformance();
    }
    
    m_collective_performance = 0;
    for (unsigned int i = 0; i < m_workers.size(); i++)
    {
        m_workers[i]->Wait();
        m_workers[i]->PrintInformation();
        m_collective_performance += m_workers[i]->m_performance;
    }
}

void WorkerPool::PrepareForNewTask(const int from, const int to, Task *new_task)
{
    m_load_sizes.resize(m_workers.size());
    
    m_subdomain_from = from;
    m_subdomain_to = to;    
    const int subdomain_x_size = to - from;
    int remaining_x;
    
    if (m_collective_performance == 0)
    {
        /* If performance has not been measured: Divide work load evenly */
        remaining_x = subdomain_x_size % m_workers.size();
        for (unsigned int i = 0; i < m_load_sizes.size(); i++)
        {
            m_load_sizes[i] = subdomain_x_size / m_workers.size();
        }
    }
    else 
    {
        /* Divide work among workers according to their measured performance */
        remaining_x = subdomain_x_size;
        for (unsigned int i = 0; i < m_load_sizes.size(); i++)
        {
            m_load_sizes[i] = subdomain_x_size * ((float)m_workers[i]->m_performance / m_collective_performance);
            remaining_x -= m_load_sizes[i];
        }    
    }

    /* Assign the remaining work (if any). */    
    for (unsigned int i = 0; i < m_workers.size(); i++)
    {
        if (remaining_x == 0) {
            break;
        }
        m_load_sizes[i]++;
        remaining_x--;
    }

    /* Prepare all the Workers in the pool for the new Task. */
    bool has_bottom_subdomain = (from == 0);
    int worker_from = 0;
    int worker_to = 0;        
    for (unsigned int i = 0; i < m_workers.size(); i++)
    {
        PrintWorkerLoad(i, m_load_sizes[i], new_task->m_num_x);
        worker_to += m_load_sizes[i];
        m_workers[i]->PrepareForNewTask(has_bottom_subdomain, worker_from, worker_to, new_task);
        has_bottom_subdomain = false;
        worker_from = worker_to;        
    }
    
    for (unsigned int i = 0; i < m_workers.size(); i++) {
        m_workers[i]->Wait();
    }    
}

void WorkerPool::PrintWorkerLoad(const unsigned int worker_id, const int subproblem_size, const int problem_size)
{
    if (g_output == ERROR) {
        return;
    }

    const float percentage = ((float)subproblem_size / problem_size) * 100;
    printf("Worker %d gets %.2f%% of the FDTD domain.\n", worker_id, percentage);    
    
    if (worker_id == m_workers.size() - 1) {
        printf("\n");
    }
}

void WorkerPool::UpdateH()
{
    for (unsigned int i = 0; i < m_workers.size(); i++) {        
        m_workers[i]->UpdateH();    
    }
                
    for (unsigned int i = 0; i < m_workers.size(); i++) {
        m_workers[i]->Wait();
    }
}

void WorkerPool::UpdateE()
{         
    for (unsigned int i = 0; i < m_workers.size(); i++) {        
        m_workers[i]->UpdateE();                
    }
                
    for (unsigned int i = 0; i < m_workers.size(); i++) {
        m_workers[i]->Wait();
    }   
}

void WorkerPool::SourceExcitation(const int time_step, const int i_source)
{    
    if (m_subdomain_from <= i_source && i_source < m_subdomain_to)
    {
        const int local_i_source = i_source - m_subdomain_from;
        for (unsigned int i = 0; i < m_workers.size(); i++) 
        {        
            if (m_workers[i]->m_from <= local_i_source && local_i_source < m_workers[i]->m_to)
            {
                m_workers[i]->SourceExcitation(time_step, local_i_source);
                m_workers[i]->Wait();
                break;
            }
        }
    }
}

void WorkerPool::FinalizeTask()
{
    for (unsigned int i = 0; i < m_workers.size(); i++) {
        m_workers[i]->FinalizeTask();
    }
    
    for (unsigned int i = 0; i < m_workers.size(); i++) {
        m_workers[i]->Wait();
    }
}

void WorkerPool::KillAllWorkers()
{
    for (unsigned int i = 0; i < m_workers.size(); i++)
    {
        m_workers[i]->Stop();
        delete m_workers[i];
    }
    
    m_workers.clear();
    m_load_sizes.clear();
}
