#ifndef WORKERPOOL_H_
#define WORKERPOOL_H_

#include <vector>

class Task;
class Worker;

/// A helper class used to interact with a set of Worker objects.
/*! This class is only meant to simplify interactions with multiple workers at the same time by providing a set of functions corresponding to the Worker class’ public functions. There should generally only exist one instance of this class per Node. It has however not been made all static like the Node class, since future development might require several instances per node. This class will internally handle all the needed synchronizations between workers by calling Worker::Wait() when necessary. */
class WorkerPool
{
public:
    /// Adds a new Worker to the #m_workers vector, nothing else.
    void AddWorker(Worker *new_worker);
    
    /// Calls the Worker::MeasurePerformance() function for all the workers in #m_workers.
    /*! This function will measure the performance of all the workers in the pool by calling the Worker::MeasurePerformance() function for all the workers in #m_workers. It will also set the #m_collective_performance, so that it contains the collective performance of all the workers in the WorkerPool. All the workers will be synchronized by calls to Worker::Wait(), before returning from this function. */
    void MeasurePerformance();
    
    /// Calls the Worker::PrepareForNewTask() function for all the workers in #m_workers.
    /*! This function will prepare all the workers in #m_workers for the new Task by calling Worker::PrepareForNewTask() on each of them. This function must be called before UpdateH(), UpdateE(), SourceExcitation() or FinalizeTask(). FinalizeTask() is the counterpart to this function, and should be called after the task is considered done. Calling this function will divide the task's domain into a number of sub-domains equal to the number of workers in the #m_workers vector. The size of the sub-domains is determined by the measured performance of each of the workers after a call to MeasurePerformance(). If MeasurePerformance() has not been called prior to calling this function, the domain will be divided evenly among the workers. All the workers will be synchronized by calling Worker::Wait(), before returning from this function.
     *  @param[in] from,to Specifies the range of the original domain which the WorkerPool should work on. If the original domain has been divided among multiple nodes, this range should specify the range of the sub-domain assigned to the node who owns this instance of the WorkerPool class. 
     *  @param[in] new_task A pointer to the new Task which the workers will prepare for. Every worker will store a copy of this pointer. Calling this function alone will not make any changes to the task instance. */
    void PrepareForNewTask(const int from, const int to, Task *new_task);
    
    /// Calls the Worker::UpdateH() function for all the workers in #m_workers.
    /*! This function will do one update of the magnetic fields using the Task and range previously set by a call to the PrepareForNewTask() function. Should be called once for every Task::m_num_time_steps. All the workers will be synchronized by calling Worker::Wait(), before returning from this function. */
    void UpdateH();
    
    /// Calls the Worker::UpdateE() function for all the workers in #m_workers.
    /*! This function will do one update of the electric fields using the Task and range previously set by a call to the PrepareForNewTask() function. Should be called once for every Task::m_num_time_steps. All the workers will be synchronized by calling Worker::Wait(), before returning from this function. */
    void UpdateE();
    
    /// Calls the Worker::SourceExcitation() function for the Worker who has the point source in its sub-domain.
    /*! This function will do a check to see if the \p i_source exists in the range previously set by calling the PrepareForNewTask() function. If the \p i_source exists in the range, the worker who is responsible for the sub-domain containing the point source will be found and signaled. The worker will then update the source point by calling Worker::SourceExcitation(). This function should be called once for every Task::m_num_time_steps. All the workers will be synchronized by calling Worker::Wait(), before returning from this function.
     *  @param[in] time_step The current time-step. Included because the algorithm calculating the new point source needs it.
     *  @param[in] i_source The coordinate of the source point in the X-dimension. Needed in order to find the worker responsible for the sub-domain containing the source point. Because the domain is only divided along the X-dimension, the Y and Z coordinates do not need to be passed to this function. The Y and Z coordinates are however still needed, but can later be fetched from the Task object by the worker doing the actual source update. */
    void SourceExcitation(const int time_step, const int i_source);
    
    /// Calls the Worker::FinalizeTask() function for all the workers.
    /*! This function is the counterpart to the PrepareForNewTask() function, and should be called after the task is considered done. All the workers will be synchronized by calling Worker::Wait(), before returning from this function. */
    void FinalizeTask();
    
    /// Calls the Worker::Stop() function for all the workers. A cleanup function that calls Worker::Stop() on all the workers before deleting them.
    /*! Calling this function will stop all the workers, delete them, and do the necessary cleanup. It should be called at the end of the program execution in order to terminate all the threads and free resources. */
    void KillAllWorkers();
    
private:
    /// Prints the precentage of the worker's sub-domain size related to the total domain size.
    /*! This function will not print anything if #g_output is set to ERROR.
     *  @param[in] worker_id The ID of the worker.
     *  @param[in] subproblem_size The size of the sub-domain, specifically the size in the X dimension.
     *  @param[in] problem_size The size of the complete domain, specifically the size in the X dimension. */
    void PrintWorkerLoad(const unsigned int worker_id, const int subproblem_size, const int problem_size);

    std::vector<Worker*> m_workers; ///< A vector of pointers to Worker objects.
    std::vector<int> m_load_sizes; ///< A vector of all the load sizes, corresponding to #m_workers.
    int m_collective_performance; ///< The collective performance of all the workers in #m_workers.
    
    int m_subdomain_from; ///< X-coordinate of where the WorkerPool's sub-domain starts in the FDTD domain.
    int m_subdomain_to; ///< X-coordinate of where the WorkerPool's sub-domain ends in the FDTD domain.
};

#endif /* WORKERPOOL_H_ */
