#include "CpuWorker.h"

#include "../Task.h"
#include <math.h>
#include <omp.h>
#include <stdio.h>

#define Ex(I,J,K) ex[yz_offset_e*(I) + (nz+1)*(J) + (K)]
#define Ey(I,J,K) ey[yz_offset_e*(I) + (nz+1)*(J) + (K)]
#define Ez(I,J,K) ez[yz_offset_e*(I) + (nz+1)*(J) + (K)]
#define Hx(I,J,K) hx[yz_offset_h*(I) + nz*(J) + (K)]
#define Hy(I,J,K) hy[yz_offset_h*(I) + nz*(J) + (K)]
#define Hz(I,J,K) hz[yz_offset_h*(I) + nz*(J) + (K)]

CpuWorker::CpuWorker(const int number_of_threads)
    : m_number_of_threads(number_of_threads)
{
    sprintf(m_name, "CPU with %d threads", number_of_threads);
}

void CpuWorker::UpdateH_Internal()
{
    const int ny = m_task->m_num_y;
    const int nz = m_task->m_num_z;
    
    const int yz_offset_e = (ny+1) * (nz+1);
    const int yz_offset_h = ny * nz;    
    
    const my_float *ex = m_task->m_ex;
    const my_float *ey = m_task->m_ey;
    const my_float *ez = m_task->m_ez;
    my_float *hx = m_task->m_hx;
    my_float *hy = m_task->m_hy;
    my_float *hz = m_task->m_hz;        
    
    const my_float cbdx = m_task->m_Cbdx;
    const my_float cbdy = m_task->m_Cbdy;
    const my_float cbdz = m_task->m_Cbdz;
    
    int i, j, k;
    
    #pragma omp parallel for num_threads(m_number_of_threads)
    for (i=m_from; i<m_to; i++) {
        for (j=0; j<ny; j++) {
            for (k=0; k<nz; k++) {
                Hx(i,j,k) += (Ey(i,j,k+1)-Ey(i,j,k))*cbdz - (Ez(i,j+1,k)-Ez(i,j,k))*cbdy;
                Hy(i,j,k) += (Ez(i+1,j,k)-Ez(i,j,k))*cbdx - (Ex(i,j,k+1)-Ex(i,j,k))*cbdz;
                Hz(i,j,k) += (Ex(i,j+1,k)-Ex(i,j,k))*cbdy - (Ey(i+1,j,k)-Ey(i,j,k))*cbdx;
            }
        }
    }
}

void CpuWorker::UpdateE_Internal()
{    
    const int ny = m_task->m_num_y;
    const int nz = m_task->m_num_z;
    
    const int yz_offset_e = (ny+1) * (nz+1);
    const int yz_offset_h = ny * nz;        
    
    my_float *ex = m_task->m_ex;
    my_float *ey = m_task->m_ey;
    my_float *ez = m_task->m_ez;
    const my_float *hx = m_task->m_hx;
    const my_float *hy = m_task->m_hy;
    const my_float *hz = m_task->m_hz;        
    
    const my_float dbdx = m_task->m_Dbdx;
    const my_float dbdy = m_task->m_Dbdy;
    const my_float dbdz = m_task->m_Dbdz;

    const int from = (m_has_bottom_subdomain == true) ? 1 : m_from;
    
    int i, j, k;

    #pragma omp parallel num_threads(m_number_of_threads)
    {
        if (m_has_bottom_subdomain == true)
        {
            #pragma omp for
            for (j=1; j<ny; j++) {
                for (k=1; k<nz; k++) {
                    Ex(0,j,k) += (Hz(0,j,k)-Hz(0,j-1,k))*dbdy + (Hy(0,j,k-1)-Hy(0,j,k))*dbdz;
                }
            }
        }
    
        #pragma omp for
        for (i=from; i<m_to; i++) {
            for (k=1; k<nz; k++) {
                Ey(i,0,k) += (Hx(i,0,k)-Hx(i,0,k-1))*dbdz + (Hz(i-1,0,k)-Hz(i,0,k))*dbdx;
            }
        }

        #pragma omp for
        for (i=from; i<m_to; i++) {
            for (j=1; j<ny; j++) {
                Ez(i,j,0) += (Hy(i,j,0)-Hy(i-1,j,0))*dbdx + (Hx(i,j-1,0)-Hx(i,j,0))*dbdy;
            }
        }

        #pragma omp for
        for (i=from; i<m_to; i++) {
            for (j=1; j<ny; j++) {
                for (k=1; k<nz; k++) {
                    Ex(i,j,k) += (Hz(i,j,k)-Hz(i,j-1,k))*dbdy + (Hy(i,j,k-1)-Hy(i,j,k))*dbdz;
                    Ey(i,j,k) += (Hx(i,j,k)-Hx(i,j,k-1))*dbdz + (Hz(i-1,j,k)-Hz(i,j,k))*dbdx;
                    Ez(i,j,k) += (Hy(i,j,k)-Hy(i-1,j,k))*dbdx + (Hx(i,j-1,k)-Hx(i,j,k))*dbdy;
                }
            }
        }
    }
}

void CpuWorker::SourceExcitationInternal()
{
    const int ny = m_task->m_num_y;
    const int nz = m_task->m_num_z;    
    const int yz_offset_e = (ny+1) * (nz+1);

    my_float *ez = m_task->m_ez;

    const my_float tk = 2.0e-9;
    const my_float temp = m_current_time_step*m_task->m_dt - 3*tk;        

    #ifdef use_float     
    Ez(m_i_source, m_task->m_js, m_task->m_ks) += -m_task->m_p_source_factor * temp/(tk*tk) * expf(-temp*temp/(tk*tk));
    #else
    Ez(m_i_source, m_task->m_js, m_task->m_ks) += -m_task->m_p_source_factor * temp/(tk*tk) * exp(-temp*temp/(tk*tk));
    #endif
}
