#ifndef CPUWORKER_H_
#define CPUWORKER_H_

#include "Worker.h"

class Task;

/// A Worker sub-class specialized for doing FDTD on a multi-core CPU.
/*! A sub-class of the abstract Worker class. Contains functions for updating the electromagnetic field values using a multi-core CPU. */
class CpuWorker: public Worker {
public:
    /// Constructor
    /*! Will create a CpuWorker instance using the specified number of OpenMP threads.
    @param[in] number_of_threads An optional parameter which specifies the number of OpenMP threads to use. */
    CpuWorker(const int number_of_threads = 1);
    
private:
    /// Updates the magnetic field values in the specified range: #m_from to #m_to, in the FDTD domain of #m_task.
    /*! This function will update the magnetic field values of the sub-domain associated with the CpuWorker. The updating is accelerated on multi-core systems by using OpenMP. */       
    void UpdateH_Internal();
    
    /// Updates the electric field values in the specified range: #m_from to #m_to, in the FDTD domain of #m_task.
    /*! This function will update the electric field values of the sub-domain associated with the CpuWorker. The updating is accelerated on multi-core systems by using OpenMP. */
    void UpdateE_Internal();
    
    /// Updates the source point (#m_i_source) in the FDTD domain of #m_task.
    /*! This function will update the source point of the domain. If this function is called it is already assumed that the worker is responsible for the sub-domain containing the source point. */
    void SourceExcitationInternal();
    
    const int m_number_of_threads; ///< The number of OpenMP threads used during executions.
};

#endif /* CPUWORKER_H_ */
