#include "GpuWorker.h"

#include "../Task.h"
#include <string.h>

#define ExC(I,J,K) device_ex[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define EyC(I,J,K) device_ey[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define EzC(I,J,K) device_ez[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define HxC(I,J,K) device_hx[ny*nz*(I) + nz*(J) + (K)]
#define HyC(I,J,K) device_hy[ny*nz*(I) + nz*(J) + (K)]
#define HzC(I,J,K) device_hz[ny*nz*(I) + nz*(J) + (K)]

namespace GpuWorkerKernels
{
__global__ void UpdateH_cuda(const int nx, const int ny, const int nz,
        const my_float Cbdx, const my_float Cbdy, const my_float Cbdz,
        const my_float *device_ex, const my_float *device_ey, const my_float *device_ez,
        my_float *device_hx, my_float *device_hy, my_float *device_hz);
__global__ void UpdateE_cuda(const bool is_bottom_subdomain, const int nx, const int ny, const int nz,
        const my_float Dbdx, const my_float Dbdy, const my_float Dbdz,
        const my_float *device_hx, const my_float *device_hy, const my_float *device_hz,
        my_float *device_ex, my_float *device_ey, my_float *device_ez);
__global__ void CleanupE_yz_cuda(const int ny, const int nz,
        const my_float Dbdy, const my_float Dbdz, const my_float *device_hy, const my_float 
        *device_hz, my_float *device_ex);
__global__ void CleanupE_xz_cuda(const bool is_bottom_subdomain, const int nx, const int ny, const int nz,
        const my_float Dbdx, const my_float Dbdz, const my_float *device_hx, const my_float *device_hz, 
        my_float *device_ey);
__global__ void CleanupE_xy_cuda(const bool is_bottom_subdomain, const int nx, const int ny, const int nz,
        const my_float Dbdx, const my_float Dbdy, const my_float *device_hx, const my_float *device_hy, 
        my_float *device_ez);
__global__ void ApplySource(const int ts, const int is, const int js, const int ks, const int ny, const int nz,
        const my_float dt, const my_float p_source_factor, my_float *device_ez);
}

int GpuWorker::s_number_of_devices = -1;
int GpuWorker::s_next_device_number;

GpuWorker::GpuWorker()
    : m_device_number(s_next_device_number++)
{
    cudaDeviceProp properties;
    cudaGetDeviceProperties(&properties, m_device_number);
    m_block_size = properties.maxThreadsPerBlock;
    strcpy(m_name, properties.name);

    if (2 <= properties.major) 
    {
        /* Only devices with compute capability 2.0 or newer have L1 cache */
        cudaSetDevice(m_device_number);
        cudaDeviceSetCacheConfig(cudaFuncCachePreferL1); 
    }
}

int GpuWorker::GetNumberOfDevices()
{
    if (s_number_of_devices == -1) {
        cudaGetDeviceCount(&s_number_of_devices);
    }   
    
    return s_number_of_devices;
}

void GpuWorker::PrepareForNewTaskInternal()
{                    
    cudaSetDevice(m_device_number);

    const int yz_offset_e = (m_task->m_num_y+1) * (m_task->m_num_z+1);
    const int yz_offset_h = m_task->m_num_y * m_task->m_num_z;
    const int e_size = (m_local_x_size+1) * yz_offset_e * sizeof(my_float);
    const int h_size = m_local_x_size * yz_offset_h * sizeof(my_float);
    
    /* Unless it has the lowest subdomain, the device needs to keep a copy of 
     * the upper H-border of the underlying Worker's subdomain in memory */        
    const int extra_border_size = (m_has_bottom_subdomain == true) ? 0 : yz_offset_h * sizeof(my_float);    

    /* Allocate memory on the device */
    cudaMalloc((void**)&device_ex, e_size);
    cudaMalloc((void**)&device_ey, e_size);
    cudaMalloc((void**)&device_ez, e_size);    
    cudaMalloc((void**)&device_hx, h_size);
    cudaMalloc((void**)&device_hy, h_size + extra_border_size);
    cudaMalloc((void**)&device_hz, h_size + extra_border_size);

    if (m_has_bottom_subdomain == false)
    {
        /* Offset the starting indexes since the arrays have an additional border at the start */
        device_hy = &device_hy[yz_offset_h];
        device_hz = &device_hz[yz_offset_h];
    }

    const my_float *host_ex = m_task->m_ex;
    const my_float *host_ey = m_task->m_ey;
    const my_float *host_ez = m_task->m_ez;
    const my_float *host_hx = m_task->m_hx;
    const my_float *host_hy = m_task->m_hy;
    const my_float *host_hz = m_task->m_hz;

    /* Load entire subdomain into device memory */
    cudaMemcpy(device_ex, &host_ex[m_from*yz_offset_e], e_size, cudaMemcpyHostToDevice);                         
    cudaMemcpy(device_ey, &host_ey[m_from*yz_offset_e], e_size, cudaMemcpyHostToDevice);
    cudaMemcpy(device_ez, &host_ez[m_from*yz_offset_e], e_size, cudaMemcpyHostToDevice);     
    cudaMemcpy(device_hx, &host_hx[m_from*yz_offset_h], h_size, cudaMemcpyHostToDevice);                         
    cudaMemcpy(device_hy, &host_hy[m_from*yz_offset_h], h_size, cudaMemcpyHostToDevice);
    cudaMemcpy(device_hz, &host_hz[m_from*yz_offset_h], h_size, cudaMemcpyHostToDevice);    
}

void GpuWorker::UpdateE_Internal()
{
    cudaSetDevice(m_device_number);

    const int ny = m_task->m_num_y;
    const int nz = m_task->m_num_z;

    if (g_single_worker == false && m_has_bottom_subdomain == false)
    {
        /* Load lower H-fields borders into device memory */
        const my_float *host_hy = m_task->m_hy;
        const my_float *host_hz = m_task->m_hz;
        const int yz_offset_h = ny * nz;
        const int h_border_size = yz_offset_h * sizeof(my_float);
        cudaMemcpy(&device_hy[-yz_offset_h], &host_hy[(m_from-1)*yz_offset_h], h_border_size, cudaMemcpyHostToDevice);
        cudaMemcpy(&device_hz[-yz_offset_h], &host_hz[(m_from-1)*yz_offset_h], h_border_size, cudaMemcpyHostToDevice);
    }

    const int grid_yz = ceil(sqrt(ny*nz / (float)m_block_size));
    const int grid_xz = ceil(sqrt(m_local_x_size*nz / (float)m_block_size));
    const int grid_xy = ceil(sqrt(m_local_x_size*ny / (float)m_block_size));

    const dim3 dimGrid_yz(grid_yz, grid_yz);
    const dim3 dimGrid_xz(grid_xz, grid_xz);
    const dim3 dimGrid_xy(grid_xy, grid_xy);
        
    if (m_has_bottom_subdomain == true)
    {       
        GpuWorkerKernels::CleanupE_yz_cuda<<<dimGrid_yz, m_block_size>>>(
            ny, nz, m_task->m_Dbdy, m_task->m_Dbdz, device_hy, device_hz, device_ex);       
    }    
       
    GpuWorkerKernels::CleanupE_xz_cuda<<<dimGrid_xz, m_block_size>>>(
        m_has_bottom_subdomain, m_local_x_size, ny, nz, m_task->m_Dbdx, m_task->m_Dbdz, device_hx, device_hz, device_ey);
    GpuWorkerKernels::CleanupE_xy_cuda<<<dimGrid_xy, m_block_size>>>(
        m_has_bottom_subdomain, m_local_x_size, ny, nz, m_task->m_Dbdx, m_task->m_Dbdy, device_hx, device_hy, device_ez);

    GpuWorkerKernels::UpdateE_cuda<<<dimGrid_yz, m_block_size>>>(
        m_has_bottom_subdomain, m_local_x_size, ny, nz, m_task->m_Dbdx, m_task->m_Dbdy, m_task->m_Dbdz,
        device_hx, device_hy, device_hz, device_ex, device_ey, device_ez);   
      
    if (g_single_worker == false && m_has_bottom_subdomain == false)
    {
        /* Un-load lower E-fields borders onto host memory */
        my_float *host_ey = m_task->m_ey;
        my_float *host_ez = m_task->m_ez;
        const int yz_offset_e = (ny+1) * (nz+1);
        const int e_border_size = yz_offset_e * sizeof(my_float);
        cudaMemcpy(&host_ey[m_from*yz_offset_e], device_ey, e_border_size, cudaMemcpyDeviceToHost);
        cudaMemcpy(&host_ez[m_from*yz_offset_e], device_ez, e_border_size, cudaMemcpyDeviceToHost);
    }
}

void GpuWorker::UpdateH_Internal()
{          
    cudaSetDevice(m_device_number);

    const int ny = m_task->m_num_y;
    const int nz = m_task->m_num_z;             
    
    if (g_single_worker == false)
    {
        /* Load upper E-fields borders into device memory */
        const my_float *host_ey = m_task->m_ey;
        const my_float *host_ez = m_task->m_ez; 
        const int yz_offset_e = (ny+1) * (nz+1);  
        const int e_border_size = yz_offset_e * sizeof(my_float);
        cudaMemcpy(&device_ey[m_local_x_size*yz_offset_e], &host_ey[m_to*yz_offset_e], e_border_size, cudaMemcpyHostToDevice);
        cudaMemcpy(&device_ez[m_local_x_size*yz_offset_e], &host_ez[m_to*yz_offset_e], e_border_size, cudaMemcpyHostToDevice);     
    }
    
    const int yz_offset_h = ny * nz;
    
    const int grid_yz = ceil(sqrt(yz_offset_h / (float)m_block_size));
    const dim3 dim_grid_yz(grid_yz, grid_yz);    
    
    GpuWorkerKernels::UpdateH_cuda<<<dim_grid_yz, m_block_size>>>(
        m_local_x_size, ny, nz, m_task->m_Cbdx, m_task->m_Cbdy, m_task->m_Cbdz,
        device_ex, device_ey, device_ez, device_hx, device_hy, device_hz);
    
    if (g_single_worker == false)
    {
        /* Un-load upper H-fields borders onto host memory */
        my_float *host_hy = m_task->m_hy;
        my_float *host_hz = m_task->m_hz;
        const int h_border_size = yz_offset_h * sizeof(my_float);
        cudaMemcpy(&host_hy[(m_to-1)*yz_offset_h], &device_hy[(m_local_x_size-1)*yz_offset_h], h_border_size, cudaMemcpyDeviceToHost);
        cudaMemcpy(&host_hz[(m_to-1)*yz_offset_h], &device_hz[(m_local_x_size-1)*yz_offset_h], h_border_size, cudaMemcpyDeviceToHost);    
    }    
}

void GpuWorker::SourceExcitationInternal()
{
    cudaSetDevice(m_device_number);

    m_i_source -= m_from;
    
    GpuWorkerKernels::ApplySource<<<1, 1>>>(
        m_current_time_step, m_i_source, m_task->m_js, m_task->m_ks, 
        m_task->m_num_y, m_task->m_num_z, m_task->m_dt, m_task->m_p_source_factor, device_ez);
}

void GpuWorker::FinalizeTaskInternal()
{
    cudaSetDevice(m_device_number);

    my_float *host_ex = m_task->m_ex;
    my_float *host_ey = m_task->m_ey;
    my_float *host_ez = m_task->m_ez;
    my_float *host_hx = m_task->m_hx;
    my_float *host_hy = m_task->m_hy;
    my_float *host_hz = m_task->m_hz;   

    /* Un-load entire E-subdomain onto host memory */
    const int yz_offset_e = (m_task->m_num_y+1) * (m_task->m_num_z+1);
    const int e_size = (m_local_x_size+1) * yz_offset_e * sizeof(my_float);
    cudaMemcpy(&host_ex[m_from*yz_offset_e], device_ex, e_size, cudaMemcpyDeviceToHost);
    cudaMemcpy(&host_ey[m_from*yz_offset_e], device_ey, e_size, cudaMemcpyDeviceToHost);
    cudaMemcpy(&host_ez[m_from*yz_offset_e], device_ez, e_size, cudaMemcpyDeviceToHost);

    /* Un-load entire H-subdomain onto host memory */
    const int yz_offset_h = m_task->m_num_y * m_task->m_num_z;  
    const int h_size = m_local_x_size * yz_offset_h * sizeof(my_float);
    cudaMemcpy(&host_hx[m_from*yz_offset_h], device_hx, h_size, cudaMemcpyDeviceToHost);
    cudaMemcpy(&host_hy[m_from*yz_offset_h], device_hy, h_size, cudaMemcpyDeviceToHost);
    cudaMemcpy(&host_hz[m_from*yz_offset_h], device_hz, h_size, cudaMemcpyDeviceToHost);                 
            
    if (m_has_bottom_subdomain == false)
    {   
        /* Reset the device's array pointers */
        device_hy = &device_hy[-yz_offset_h];
        device_hz = &device_hz[-yz_offset_h];
    }           
            
    /* Free the device's memory since the task has been completed */
    cudaFree(device_ex);
    cudaFree(device_ey);
    cudaFree(device_ez);
    cudaFree(device_hx);
    cudaFree(device_hy);
    cudaFree(device_hz);    
}

namespace GpuWorkerKernels
{

__global__ void UpdateH_cuda(const int nx, const int ny, const int nz,
        const my_float Cbdx, const my_float Cbdy, const my_float Cbdz,
        const my_float *device_ex, const my_float *device_ey, const my_float *device_ez,
        my_float *device_hx, my_float *device_hy, my_float *device_hz)
{
    const int id = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + threadIdx.x;

    const int y = id/nz;
    const int z = id%nz;

    if (ny <= y || nz <= z) {
        return;
    }    

    for (int x=0; x<nx; x++)
    {
        HxC(x,y,z) += (EyC(x,y,z+1)-EyC(x, y ,z))*Cbdz + (EzC(x,y, z )-EzC(x,y+1,z))*Cbdy;
        HyC(x,y,z) += (EzC(x+1,y,z)-EzC(x,y, z ))*Cbdx + (ExC( x ,y,z)-ExC(x,y,z+1))*Cbdz;
        HzC(x,y,z) += (ExC(x,y+1,z)-ExC( x ,y,z))*Cbdy + (EyC(x, y ,z)-EyC(x+1,y,z))*Cbdx;
    }
}

__global__ void CleanupE_yz_cuda(const int ny, const int nz,
        const my_float Dbdy, const my_float Dbdz, 
        const my_float *device_hy, const my_float *device_hz, my_float *device_ex)
{
	const int id = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + threadIdx.x;

    const int y = id/nz + 1;
    const int z = id%nz + 1;

    if (y < ny && z < nz) {
        ExC(0, y, z) += (HzC(0, y, z)-HzC(0, y-1, z))*Dbdy + (HyC(0, y, z-1)-HyC(0, y, z))*Dbdz;
    }
}

__global__ void CleanupE_xz_cuda(const bool is_bottom_subdomain, const int nx, const int ny, const int nz,
        const my_float Dbdx, const my_float Dbdz, 
        const my_float *device_hx, const my_float *device_hz, my_float *device_ey)
{
	const int id = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + threadIdx.x;

    const int x = id/nz + is_bottom_subdomain;
    const int z = id%nz + 1;

    if (x < nx && z < nz) {
        EyC(x, 0, z) += (HxC(x, 0, z)-HxC(x, 0, z-1))*Dbdz + (HzC(x-1, 0, z)-HzC(x, 0, z))*Dbdx;
    }
}

__global__ void CleanupE_xy_cuda(const bool is_bottom_subdomain, const int nx, const int ny, const int nz,
        const my_float Dbdx, const my_float Dbdy, 
        const my_float *device_hx, const my_float *device_hy, my_float *device_ez)
{
	const int id = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + threadIdx.x;

    const int x = id/ny + is_bottom_subdomain;
    const int y = id%ny + 1;

    if (x < nx && y < ny) {
        EzC(x, y, 0) += (HyC(x, y, 0)-HyC(x-1, y, 0))*Dbdx + (HxC(x, y-1, 0)-HxC(x, y, 0))*Dbdy;
    }
}

__global__ void UpdateE_cuda(const bool is_bottom_subdomain, const int nx, const int ny, const int nz,
        const my_float Dbdx, const my_float Dbdy, const my_float Dbdz,
        const my_float *device_hx, const my_float *device_hy, const my_float *device_hz,
        my_float *device_ex, my_float *device_ey, my_float *device_ez)
{
	const int id = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + threadIdx.x;

    const int y = id/nz + 1;
    const int z = id%nz + 1;

    if (ny <= y || nz <= z) {
        return;
    }

    for (int x=is_bottom_subdomain; x<nx; x++)
    {
    		__syncthreads();
        ExC(x,y,z) += (HzC(x,y,z)-HzC(x,y-1,z))*Dbdy + (HyC(x,y,z-1)-HyC(x,y,z))*Dbdz;
        EyC(x,y,z) += (HxC(x,y,z)-HxC(x,y,z-1))*Dbdz + (HzC(x-1,y,z)-HzC(x,y,z))*Dbdx;
        EzC(x,y,z) += (HyC(x,y,z)-HyC(x-1,y,z))*Dbdx + (HxC(x,y-1,z)-HxC(x,y,z))*Dbdy;
    }
}

__global__ void ApplySource(const int ts, const int is, const int js, const int ks, 
        const int ny, const int nz,
        const my_float dt, const my_float p_source_factor, my_float *device_ez)
{
    const my_float Tk = 2.0e-9;
    const my_float temp = ts*dt - 3*Tk;
    
    #ifdef use_float
    EzC(is,js,ks) += -p_source_factor * temp/(Tk*Tk) * expf(-temp*temp/(Tk*Tk));
    #else
    EzC(is,js,ks) += -p_source_factor * temp/(Tk*Tk) * exp(-temp*temp/(Tk*Tk));
    #endif
}

}

