#ifndef GPUWORKER_H_
#define GPUWORKER_H_

#include "../globals.h"
#include "Worker.h"

class Task;

/// A Worker sub-class specialized for doing FDTD on a CUDA-capable GPU.
/*! A sub-class of the abstract Worker class which contains functions for updating the electromagnetic field values using a CUDA-capable GPU. */
class GpuWorker: public Worker {
public:
    /// Constructor
    /*! Will set #m_device_number and m_block_size when called, and increment #s_next_device_number. */
    GpuWorker();
    
    /// Returns the number of avaliable CUDA-capable GPUs.
    /*! Will update #s_number_of_devices the first time it's called. */
    static int GetNumberOfDevices();        
    
private:
    /// Preparations for a new Task, executed by the internal thread.
    /*! This function will allocate the necessary memory on the GPU, before copying the electromagnetic field values into the GPU’s device memory. */
    void PrepareForNewTaskInternal();
    
    /// Updates the magnetic field values in the specified range: #m_from to #m_to, in the FDTD domain of #m_task.
    /*! This function will copy the upper electrical field borders into device memory, before updating the magnetic field values by calling a CUDA kernel. The upper magnetic field borders will afterwards be copied to the system memory. Border exchanges will not be done if only a single worker is used. */
    void UpdateH_Internal();
    
    /// Updates the electric field values in the specified range: #m_from to #m_to, in the FDTD domain of #m_task.
    /*! This function will copy the lower magnetic field borders into device memory, before updating the electric field values by calling four CUDA kernels. The lower electric field borders will afterwards be copied to the system memory. Border exchanges will not be done if only a single worker is used, or if the worker is responsible for the bottom sub-domain (in relation to how the domain is divided among nodes and workers). */    
    void UpdateE_Internal();
    
    /// Updates the source point (#m_i_source) in the FDTD domain of #m_task.
    /*! This function will call one CUDA kernel using a single CUDA core to update the source point of the domain. If this function is called it is already assumed that the worker is responsible for the sub-domain containing the source point.*/
    void SourceExcitationInternal();
    
    /// Finalization of the Task done by the internal thread.
    /*! This function will copy all the electromagnetic field values of the sub-domain to system memory, and then remove (free) them from device memory. */
    void FinalizeTaskInternal();
    
    static int s_number_of_devices; ///< The number of avaliable CUDA-capable GPUs
    static int s_next_device_number; ///< The number of the next unused device.
    const int m_device_number; ///< The device number of the GPU associated with this GpuWorker.
    int m_block_size; ///< The CUDA block size used during kernel calls.
        
    my_float *device_hx, *device_hy, *device_hz;
    my_float *device_ex, *device_ey, *device_ez;
};

#endif /* GPUWORKER_H_ */
