#include "Worker.h"

#include "../Task.h"
#include <stdio.h>
#include <stdlib.h>

/* A static counter used to generate IDs */
int Worker::s_next_id = 0;

Worker::Worker()
    : m_id(s_next_id++), m_performance(-1.0), m_state(IDLE)    
{
    pthread_mutex_init(&m_mutex, NULL);
    pthread_cond_init(&m_condition, NULL);

    const int rc = pthread_create(&m_thread, NULL, InternalThreadEntryFunc, this);
    if (rc) {
        printf("ERROR: pthread_create() failed with return code = %d\n", rc);
        exit(EXIT_FAILURE);
    }    
}

Worker::~Worker()
{
    pthread_mutex_destroy(&m_mutex);
    pthread_cond_destroy(&m_condition);
}

void Worker::PrintInformation()
{
    if (g_output == ERROR) {
        return;
    }
    
    printf("Worker %d\n"
        " - Name: %s\n"
        " - Measured MFLOPs/s: %d\n\n",
        m_id, m_name, m_performance);
}

void Worker::PrepareForNewTask(const bool has_bottom_subdomain, const int from, const int to, Task *new_task)
{    
    m_has_bottom_subdomain = has_bottom_subdomain;    
    m_from = from;
    m_to = to;
    m_task = new_task;
    m_local_x_size = to - from;
    
    PrepareForNewTaskInternal();
}

void Worker::FinalizeTask()
{
    FinalizeTaskInternal();
}

void Worker::PrepareForNewTaskInternal()
{}

void Worker::FinalizeTaskInternal() 
{}

void Worker::UpdateH()
{    
    pthread_mutex_lock(&m_mutex);             
    m_state = UPDATE_H;    
    pthread_cond_signal(&m_condition);
    pthread_mutex_unlock(&m_mutex);
}

void Worker::UpdateE()
{
    pthread_mutex_lock(&m_mutex);        
    m_state = UPDATE_E;        
    pthread_cond_signal(&m_condition);
    pthread_mutex_unlock(&m_mutex);
}

void Worker::SourceExcitation(const int time_step, const int i_source)
{    
    pthread_mutex_lock(&m_mutex);
    m_current_time_step = time_step;         
    m_i_source = i_source;
    m_state = SOURCE_EXCITATION;
    pthread_cond_signal(&m_condition);
    pthread_mutex_unlock(&m_mutex);
}

void Worker::MeasurePerformance()
{
    pthread_mutex_lock(&m_mutex);    
    m_state = BENCHMARK;    
    pthread_cond_signal(&m_condition);
    pthread_mutex_unlock(&m_mutex);    
}

void Worker::MeasurePerformanceInternal()
{
    Task dummy(200, 100, 100, 100, 1.0, 1.0, 1.0);
    dummy.InitializeFieldComponents();    

    PrepareForNewTask(true, 0, dummy.m_num_x, &dummy);
    m_i_source = dummy.m_is;

    dummy.m_computeTimer.Start();
    for (int ts = 1; ts <= dummy.m_num_time_steps; ts++)
    {
        m_current_time_step++;
        
        UpdateE_Internal();
        SourceExcitationInternal();        
        UpdateH_Internal();
    }
    dummy.m_computeTimer.Stop();

    const double seconds = dummy.m_computeTimer.GetTime();
    const double mflops = dummy.m_flops_work / (1000*1000);    
    m_performance = static_cast<int>(mflops / seconds);
    
    FinalizeTask();
    dummy.CleanUp();        
}

void* Worker::InternalThreadEntryFunc(void *this_worker)
{
    ((Worker*)this_worker)->InternalThreadEntry();    
    pthread_exit(NULL);
}

void Worker::InternalThreadEntry()
{
    pthread_mutex_lock(&m_mutex);
    while (true)
    {                                                
        switch (m_state)
        {
            case UPDATE_H:
                UpdateH_Internal();
                break;
            case UPDATE_E:
                UpdateE_Internal();
                break;
            case SOURCE_EXCITATION:
                SourceExcitationInternal();
                break;
            case BENCHMARK:
                MeasurePerformanceInternal();
                break;
            case STOP:
                m_state = IDLE;
                pthread_cond_signal(&m_condition);    
                goto stop;
            case IDLE:
                break;            
        }
        
        m_state = IDLE;
        pthread_cond_signal(&m_condition);        
        pthread_cond_wait(&m_condition, &m_mutex);                            
    }
    stop:
    pthread_mutex_unlock(&m_mutex);
}

void Worker::Wait()
{
    pthread_mutex_lock(&m_mutex);
    while (m_state != IDLE)
        pthread_cond_wait(&m_condition, &m_mutex);
    pthread_mutex_unlock(&m_mutex);    
}

void Worker::Stop()
{
    pthread_mutex_lock(&m_mutex);
    m_state = STOP;
    pthread_cond_signal(&m_condition);
    pthread_mutex_unlock(&m_mutex);
        
    const int rc = pthread_join(m_thread, NULL);
    if (rc) {
        printf("ERROR: pthread_join() failed with return code = %d\n", rc);
        exit(EXIT_FAILURE);
    }
}

