#ifndef WORKER_H_
#define WORKER_H_

#include <pthread.h>

class Task;

/// An abstract class for workers.
/*! This class serves as an abstract class that can be used to create workers, which are specialized or optimized for a particular execution unit. It will also incorporate all the threading functionality needed to execute multiple workers simultaneously as several threads. */
class Worker
{
    friend class WorkerPool; ///< Needed because WorkerPool::SourceExcitation() needs access to #m_from and #m_to.

public:
    /// Constructor
    /*! This function will implicitly be called by the constructor of the workers which inherits this class. It will create a new POSIX thread and have it enter a wait state, where it will wait for signals from the master thread through calls to the Worker class’ public functions.*/
    Worker();
    
    /// Destructor
    /*! This destructor will free all the POSIX thread resources allocated by Worker(). The internal thread must be exited by calling: Stop(), prior to destroying the object.  */
    virtual ~Worker();
    
    /// Prints information about the worker.
    /*! This function will print the ID, the name and the measured performance of the worker. However, it will not output anything if the global variable: #g_output, is set to ERROR. */
    void PrintInformation();
    
    /// Prepares the worker for a new Task by updating the necessary member variables.
    /*! This function will prepare the worker for a new task, and must therefore be called before simulations can be done on a FDTD sub-domain. Since a call to this function only signals the internal thread before returning, a call to Wait() should follow to make sure that the internal thread has completed its execution.     
     *  @param[in] has_bottom_subdomain A flag that indicates if the sub-domain assigned to the worker is the bottom sub-domain, in relation to how the FDTD domain is divided among nodes and workers along the X-axis.
     *  @param[in] from,to Specifies the range of the node's sub-domain in which the worker's sub-domain exists.
     *  @param[in] new_task A pointer to the task which the worker should prepare for. Meaning that future calls to UpdateH(), UpdateE(), SourceExcitation() and FinalizeTask() will make changes to this task. */
    void PrepareForNewTask(const bool has_bottom_subdomain, const int from, const int to, Task *new_task);
    
    /// Signals the internal thread of the Worker that it should call UpdateH_Internal().
    /*! This function will signal the worker to update the magnetic field values of a sub-domain. Since a call to this function only signals the internal thread before returning, a call to Wait() should follow to make sure that the internal thread has completed its execution. */
    void UpdateH();
    
    /// Signals the internal thread of the Worker that it should call UpdateE_Internal().
    /*! This function will signal the worker to update the electric field values of a sub-domain. Since a call to this function only signals the internal thread before returning, a call to Wait() should follow to make sure that the internal thread has completed its execution. */    
    void UpdateE();
    
    /// Signals the internal thread of the Worker that it should call SourceExcitationInternal().
    /*! This function will signal the worker to update the source point in a domain. A check should be done prior to calling this function to make sure that sub-domain assigned to this worker actually contains the source point. Member variables will be updated with the arguments to this function, so that the thread calling: SourceExcitationInternal(), will have access to them. Since a call to this function only signals the internal thread before returning, a call to Wait() should follow to make sure that the internal thread has completed its execution.
     *  @param[in] time_step The current time-step of the FDTD simulation. Will update #m_current_time_step.
     *  @param[in] i_source The X-coordinate of the source point. Will update #m_i_source. */
    void SourceExcitation(const int time_step, const int i_source);
    
    /// Signals the internal thread of the Worker that it should call FinalizeTaskInternal().
    /*! Since a call to this function only signals the internal thread before returning, a call to Wait() should follow
     *  to make sure that the internal thread has completed its execution. */     
    void FinalizeTask();
    
    /// Signals the internal thread of the Worker that it should call MeasurePerformanceInternal().
    /*! Since a call to this function only signals the internal thread before returning, a call to Wait() should follow
     *  to make sure that the internal thread has completed its execution. */    
    void MeasurePerformance();
    
    /// Waits until the internal thread of the Worker reaches an IDLE state.
    void Wait();
    
    /// Signals the internal thread of the Worker that it should stop.
    /*! This function will wait until the internal thread completes whatever it is currently doing. It must be called before destroying the worker, so that the internal thread can exit its loop and be joined. A call to this function will only return when the internal thread has been joined, a call to Wait() is therefore not needed and will in fact create a deadlock. */
    void Stop();
    
    const int m_id; ///< An unique ID automatically generated during object construction.
    int m_performance; ///< The measured performance of the Worker in million floating-point operations per second.     

protected:
    /// Preparations for a new Task, executed by the internal thread.
    /*! This is a virtual function that does nothing and can optionally be implemented by classes inheriting this class. It has to be implemented by workers that uses execution units with dedicated memory (such as GPUs), since they need to load data prior to execution. */
    virtual void PrepareForNewTaskInternal();
    
    /// Updates the magnetic field values in the specified range: #m_from to #m_to, in the FDTD domain of #m_task.
    /*! This is a pure virtual function, meaning that it must be implemented by any class inheriting this class. */
    virtual void UpdateH_Internal()=0;
    
    /// Updates the electric field values in the specified range: #m_from to #m_to, in the FDTD domain of #m_task.
    /*! This is a pure virtual function, meaning that it must be implemented by any class inheriting this class. */    
    virtual void UpdateE_Internal()=0;
    
    /// Updates the source point (#m_i_source) in the FDTD domain of #m_task.
    /*! This is a pure virtual function, meaning that it must be implemented by any class inheriting this class. */    
    virtual void SourceExcitationInternal()=0;
    
    /// Finalization of the Task done by the internal thread.
    /*! This is a virtual function that does nothing and can optionally be implemented by classes inheriting this class. If implemented, it should clean up any resources allocated by the PrepareForNewTaskInternal() function. The function has to be implemented by workers that uses execution units with dedicated memory (such as GPUs), since they need to copy the final data to system memory at the end of the execution. */
    virtual void FinalizeTaskInternal();
    
    char m_name[256]; ///< Container for the name of the Worker.
    bool m_has_bottom_subdomain; ///< A flag that indicates if the Worker has been assigned the bottom sub-domain of the FDTD domain.
    int m_current_time_step; ///< The current time-step of the simulation. Is updated by SourceExcitation() and used by SourceExcitationInternal().
    int m_i_source; ///< The X-coordinate of the source point. Is updated by SourceExcitation() and used by SourceExcitationInternal().
    int m_local_x_size; ///< The size of the range between #m_from and #m_to.
    int m_from; ///< The X-coordinate in the node's sub-domain in which the worker's currently assigned sub-domain starts. 
    int m_to; ///< The X-coordinate in the node's sub-domain in which the worker's currently assigned sub-domain ends.
    Task *m_task; ///< Pointer to the worker's currently assigned Task.

private:
    /// The entry point of the internal thread.
    /*! This function is needed because pthread_create() cannot take a member function as a parameter. This function circumvents this limitation by being a static function that takes a pointer to a worker object as a parameter, then calls the worker object’s member function: InternalThreadEntry(), which contains the internal thread’s main loop.
     *  @param[in,out] this_worker A pointer to the worker object that owns the thread executing this function. */
    static void* InternalThreadEntryFunc(void *this_worker);
    
    /// The main loop of the internal thread.
    /*! This function contains the main loop which the internal thread will execute. The internal thread will not use polling, but rather wait (using a conditional variable) for #m_state to change. Depending on #m_state, the internal thread can execute the internal functions:  PrepareForNewTaskInternal(), MeasurePerformanceInternal(), UpdateH_Internal(), UpdateE_Internal(), SourceExcitationInternal() and FinalizeTaskInternal(). */
    void InternalThreadEntry();
    
    /// Measures the performance of the Worker and updates #m_performance accordingly.
    /*! This function will measure the performance of the worker by letting it execute a dummy task. A dummy task is both created and destroyed during a call to this function. */
    void MeasurePerformanceInternal();        
    
    static int s_next_id; ///< A static counter used to generate the unique #m_id.
    
    /// The states in which the internal thread can be in.
    enum State 
    { 
        IDLE, ///< Set by the internal thread when it has completed its current work and awaits new work.
        UPDATE_H, ///< Set by UpdateH().
        UPDATE_E, ///< Set by UpdateE().
        SOURCE_EXCITATION, ///< Set by SourceExcitation().
        BENCHMARK, ///< Set by MeasurePerformance().
        STOP ///< Set by Stop().
    };
    State m_state; ///< The state of the worker's internal thread.
        
    pthread_t m_thread; ///< The internal thread. Is initialized and started during a call to Worker().
    pthread_mutex_t m_mutex; ///< The mutex governing access to the member variables between the host and the internal thread.
    pthread_cond_t m_condition; ///< A conditional variable used for signaling the internal thread.
};

#endif /* WORKER_H_ */
