#include "main.h"

#include "Node.h"
#include "Task.h"
#include "Timer.h"
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <string>

Output g_output;

int main()
{    
    int ncpus, nomps, max_gpus;
    LoadSystemConfig(&ncpus, &nomps, &max_gpus);
    
    Node node(ncpus, nomps, max_gpus);
    
    if (node.mpi_rank == 0)
    {
        std::vector<Task*> tasks;
        int ntasks, nts, nx, ny, nz;
        my_float dx, dy, dz;
        LoadTaskConfig(&ntasks, &nts, &nx, &ny, &nz, &dx, &dy, &dz);

        for (int i = 0; i < ntasks; i++) {
            tasks.push_back( new Task(nts, nx, ny, nz, dx, dy, dz) );
        }

        for (unsigned int i = 0; i < tasks.size(); i++) 
        {    
            tasks[i]->InitializeFieldComponents();
            node.ExecuteTask(tasks[i]);
            tasks[i]->PrintResults();
            tasks[i]->CleanUp();
        }                
        node.BroadcastEndSignal();

        PrintEndResults(tasks);

        for (unsigned int i = 0; i < tasks.size(); i++) {            
            delete tasks[i];
        }    
    }
    else
    {
        node.RunAsSlave();
    }
            
    return 0;
}

void LoadSystemConfig(int *ncpus, int *nomps, int *max_ngpus)
{
    std::ifstream file("config");
    if (file.is_open() == false)
    {
        printf("ERROR: Could not open file.\n");
        exit(EXIT_FAILURE);        
    }
        
    int output_config;
    std::string line;        
    while (std::getline(file, line)) 
    {                
        line = line.substr(0, line.find("#"));
        std::istringstream inss(line.substr(line.find("=") + 1));
        if (line.find("output") != std::string::npos)
            inss >> output_config;        
        else if (line.find("cpus") != std::string::npos)
            inss >> *ncpus;
        else if (line.find("openmp") != std::string::npos)
            inss >> *nomps;
        else if (line.find("maxgpus") != std::string::npos)
            inss >> *max_ngpus;
    }
    
    g_output = static_cast<Output>(output_config); 
}

void LoadTaskConfig(int *ntasks, int *nts, int *nx, int *ny, int *nz, 
    my_float *dx, my_float *dy, my_float *dz)
{
    std::ifstream file("config");
    if (file.is_open() == false)
    {
        printf("ERROR: Could not open file.\n");
        exit(EXIT_FAILURE);        
    }

    std::string line;    
    while (std::getline(file, line)) 
    {
        line = line.substr(0, line.find("#"));
        std::istringstream inss(line.substr(line.find("=") + 1));
        if (line.find("ntasks") != std::string::npos)
            inss >> *ntasks;            
        else if (line.find("timesteps") != std::string::npos)
            inss >> *nts;
        else if (line.find("nx") != std::string::npos)
            inss >> *nx;
        else if (line.find("ny") != std::string::npos)
            inss >> *ny;
        else if (line.find("nz") != std::string::npos)
            inss >> *nz;
        else if (line.find("dx") != std::string::npos)
            inss >> *dx;
        else if (line.find("dy") != std::string::npos)
            inss >> *dy;
        else if (line.find("dz") != std::string::npos)
            inss >> *dz;                                                
    }
    
    PrintTaskConfiguration(*ntasks, *nts, *nx, *ny, *nz, *dx, *dy, *dz);
}

void PrintTaskConfiguration(const int ntasks, const int nts, 
    const int nx, const int ny, const int nz, 
    const my_float dx, const my_float dy, const my_float dz)    
{
    if (g_output == ERROR) {
        return;
    }
    
    double mflops;
    mflops =  double(ny-1)*(nz-1)*6;
    mflops += double(nx-1)*(nz-1)*6;
    mflops += double(nx-1)*(ny-1)*6;
    mflops += double(nx-1)*(ny-1)*(nz-1)*18;
    mflops += double(nx)*ny*nz*18;
    mflops *= nts;
    mflops /= (1000*1000);

    printf("Task config loaded:\n"
        " - Number of tasks: %d\n"
        " - Time steps per task: %d\n"
        " - Task dimensions (nx,ny,nz): %d, %d, %d\n"
        " - Cell dimensions (dx,dy,dz): %.3f, %.3f, %.3f\n"
        " - MFLOPs per task: %.3f\n"
        " - Total MFLOPs: %.3f\n\n",
        ntasks, nts, nx,ny,nz, dx,dy,dz, mflops, mflops*ntasks);
}

void PrintEndResults(const std::vector<Task*> &tasks)
{
    double total_mflops = 0;
    double total_setup_time = 0;    
    double total_compute_time = 0;
    double total_time = 0;  
    for (unsigned int i = 0; i < tasks.size(); i++)
    {
        total_mflops += tasks[i]->m_mflops_performance;
        total_setup_time += tasks[i]->m_setupTimer.GetTime();
        total_compute_time += tasks[i]->m_computeTimer.GetTime();
        total_time += tasks[i]->m_totalTimer.GetTime();
    }

    const double average_mflops = total_mflops / tasks.size();
    const double average_setup_time = total_setup_time / tasks.size();
    const double average_compute_time = total_compute_time / tasks.size();   

    printf("Execution completed successfully!\n"
        " - Total execution time: %.3f sec.\n"
        " - Total number of tasks: %lu\n"
        " - Average setup time per task: %.3f sec.\n"
        " - Average compute time per task: %.3f sec.\n"
        " - Average MFLOPs/s per task: %.3f\n",
    total_time, tasks.size(), average_setup_time, average_compute_time, average_mflops);
}

