#ifndef MAIN_H_
#define MAIN_H_

/*! \file main.h 
 *  @brief Helper functions for main. */

#include "globals.h"
#include <vector>

class Task;
class Worker;

/// Loads the Worker configurations from the \p config file in the root folder.
/*! This function will load the user specified configurations from the config file, which includes: The number of CPU workers, OpenMP threads per CPU worker, and the maximum allowed number of GPU workers. These values will be loaded from the file and stored in the corresponding pointers passed as arguments. This function should be used in combination with the Node class, since Node::Node() requires these values as arguments.
 *  @param[out] ncpus The number of CPU workers.
 *  @param[out] nomps The number of OpenMP threads per CPU worker.
 *  @param[out] max_ngpus The maximum allowed number of GPU workers. */
void LoadSystemConfig(int *ncpus, int *nomps, int *max_ngpus);

/// Loads the Task configurations from the \p config file in the root folder.
/*! This function will load the Task configurations from the config file and store them at the addresses pointed to by the parameters, which are intentionally equivalent to the input parameters of Task::Task(). This function should only be called by the host node, since it is responsible for creating the Tasks and distribute the workload among the slave nodes. The loaded task configurations are printed by a call to: PrintTaskConfiguration().
 *  @param[out] ntasks The number of tasks.
 *  @param[out] nts The number of time-steps.
 *  @param[out] nx,ny,nz The size of each domain.
 *  @param[out] dx,dy,dz The size of each Yee-cell in a Task. */
void LoadTaskConfig(int *ntasks, int *nts, int *nx, int *ny, int *nz, 
    my_float *dx, my_float *dy, my_float *dz);
    
/// Prints the Task configuration.
/*! Will print the Task configurations supplied as parameters. Will also calculate and print the number of floating-point operations of the Task.
 *  @param[in] ntasks The number of tasks.
 *  @param[in] nts The number of time-steps.
 *  @param[in] nx,ny,nz The size of each domain.
 *  @param[in] dx,dy,dz The size of each Yee-cell in a Task. */
void PrintTaskConfiguration(const int ntasks, const int nts, 
    const int nx, const int ny, const int nz, 
    const my_float dx, const my_float dy, const my_float dz);
    
/// Prints the end results of the program execution.
/*! Will print a summary of the program execution.
 *  @param[in] tasks Pointer to a vector containing all the completed Tasks. */  
void PrintEndResults(const std::vector<Task*> &tasks);

#endif /* MAIN_H_ */
